/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::HokuyoLaserUnit
 * @author     Fabian Paus ( fabian dot paus at kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_COMPONENT_RobotAPI_HokuyoLaserUnit_H
#define _ARMARX_COMPONENT_RobotAPI_HokuyoLaserUnit_H

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/services/tasks/PeriodicTask.h>
#include <RobotAPI/components/units/SensorActorUnit.h>
#include <RobotAPI/interface/units/LaserScannerUnit.h>

#include <urg_sensor.h>
#include <vector>

namespace armarx
{
    /**
     * @class HokuyoLaserUnitPropertyDefinitions
     * @brief
     */
    class HokuyoLaserUnitPropertyDefinitions:
        public armarx::ComponentPropertyDefinitions
    {
    public:
        HokuyoLaserUnitPropertyDefinitions(std::string prefix):
            armarx::ComponentPropertyDefinitions(prefix)
        {
            defineOptionalProperty<std::string>("LaserScannerTopicName", "LaserScans", "Name of the laser scan topic.");
            defineOptionalProperty<int>("UpdatePeriod", 25, "Update period for laser scans");
            defineOptionalProperty<float>("AngleOffset", -2.3561944902, "Offset is applied the raw angles before reporting them");
            defineOptionalProperty<std::string>("Devices", "", "List of devices in form of 'IP1,port1;IP2,port2;...'");
        }
    };

    struct HokuyoLaserScanDevice
    {
        std::string ip;
        int port = 0;
        bool connected = false;
        bool isDummy = false;
        urg_t urg;
        std::vector<long> lengthData;
    };

    /**
     * @defgroup Component-HokuyoLaserUnit HokuyoLaserUnit
     * @ingroup RobotAPI-Components
     * A description of the component HokuyoLaserUnit.
     *
     * @class HokuyoLaserUnit
     * @ingroup Component-HokuyoLaserUnit
     * @brief Brief description of class HokuyoLaserUnit.
     *
     * Detailed description of class HokuyoLaserUnit.
     */
    class HokuyoLaserUnit :
        virtual public armarx::LaserScannerUnitInterface,
        virtual public armarx::SensorActorUnit
    {
    public:
        /**
         * @see armarx::ManagedIceObject::getDefaultName()
         */
        virtual std::string getDefaultName() const
        {
            return "HokuyoLaserUnit";
        }

    protected:
        /**
         * @see armarx::ManagedIceObject::onInitComponent()
         */
        virtual void onInitComponent();

        /**
         * @see armarx::ManagedIceObject::onConnectComponent()
         */
        virtual void onConnectComponent();

        /**
         * @see armarx::ManagedIceObject::onDisconnectComponent()
         */
        virtual void onDisconnectComponent();

        /**
         * @see armarx::ManagedIceObject::onExitComponent()
         */
        virtual void onExitComponent();

        /**
         * @see PropertyUser::createPropertyDefinitions()
         */
        virtual armarx::PropertyDefinitionsPtr createPropertyDefinitions();

        std::string getReportTopicName(const Ice::Current& c) const override;

    private:
        void updateScanData();

    private:
        std::string topicName;
        LaserScannerUnitListenerPrx topic;
        int updatePeriod = 25;
        float angleOffset = 0.0f;
        std::vector<HokuyoLaserScanDevice> devices;
        PeriodicTask<HokuyoLaserUnit>::pointer_type task;
    };
}

#endif
