/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::PassThroughController
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_LIB_RobotAPI_PassThroughController_H
#define _ARMARX_LIB_RobotAPI_PassThroughController_H

#include <atomic>
#include <ArmarXCore/core/util/algorithm.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <RobotAPI/libraries/RobotRTControllers/LVL1Controller.h>
#include <RobotAPI/libraries/RobotRTControllers/RobotUnit.h>
#include <RobotAPI/interface/libraries/RTControllers/PassThroughController.h>

#include <RobotAPI/libraries/RobotRTControllers/Targets/JointPositionTarget.h>
#include <RobotAPI/libraries/RobotRTControllers/Targets/JointTorqueTarget.h>
#include <RobotAPI/libraries/RobotRTControllers/Targets/JointVelocityTarget.h>
#include <RobotAPI/libraries/RobotRTControllers/DataUnits/KinematicDataUnit.h>
#include <RobotAPI/libraries/RobotRTControllers/Targets/PlatformWheelVelocityTarget.h>

namespace armarx
{
    template <typename T>
    struct AtomicWrapper
    {
        std::atomic<T> val;

        AtomicWrapper(): val {                } {}
        AtomicWrapper(const T& v): val {v               } {}
        AtomicWrapper(const std::atomic<T>& val): val {val.load()      } {}
        AtomicWrapper(const AtomicWrapper& other): val {other.val.load()} {}

        AtomicWrapper& operator=(const AtomicWrapper& other)
        {
            val.store(other.val.load());
            return *this;
        }
    };

    template <typename TargetType>
    struct PassThroughControllerTargetTypeTraits;

    template <typename TargetType>
    class PassThroughController:
        virtual public LVL1Controller,
        virtual public PassThroughControllerInterface
    {
    public:
        using Traits = PassThroughControllerTargetTypeTraits<TargetType>;

        inline PassThroughController(LVL1ControllerDataProviderInterfacePtr prov, LVL1ControllerConfigPtr config);

        inline virtual void rtRun(const IceUtil::Time&, const IceUtil::Time&) override;
        inline virtual void rtSwapBufferAndRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration) override;
        inline virtual void rtPreActivateController() override;

        //ice interface
        inline virtual std::string getClassName(const Ice::Current& = GlobalIceCurrent) const override;
        inline virtual void setJoint(const std::string& name, Ice::Float value, const Ice::Current& = GlobalIceCurrent) override;
        inline virtual void setJoints(const StringFloatDictionary& values, const Ice::Current& = GlobalIceCurrent) override;
    protected:
        std::vector<float*> lvl0Targets;
        std::vector<const float*> jointStates;
        std::vector<AtomicWrapper<float>> iceTargets;
        std::unordered_map<std::string, std::size_t> indices;

        virtual void onInitComponent() override {}
        virtual void onConnectComponent() override {}
    };

    template <>
    struct PassThroughControllerTargetTypeTraits<JointVelocityTarget>
    {
        static float* getTargetDatamember(JointVelocityTarget& t)
        {
            return  &t.velocity;
        }
        static std::vector<const float*> getSensorValues(const KinematicDataUnitInterface* ku, const std::vector<std::string>& v)
        {
            return ku->getJointVelocities(v);
        }
        static  std::string getControlMode()
        {
            return ControlModes::VelocityMode;
        }
    };
    template <>
    struct PassThroughControllerTargetTypeTraits<JointTorqueTarget>
    {
        static float* getTargetDatamember(JointTorqueTarget& t)
        {
            return  &t.torque;
        }
        static std::vector<const float*> getSensorValues(const KinematicDataUnitInterface* ku, const std::vector<std::string>& v)
        {
            return ku->getJointTorques(v);
        }
        static  std::string getControlMode()
        {
            return ControlModes::TorqueMode;
        }
    };
    template <>
    struct PassThroughControllerTargetTypeTraits<JointPositionTarget>
    {
        static float* getTargetDatamember(JointPositionTarget& t)
        {
            return  &t.position;
        }
        static std::vector<const float*> getSensorValues(const KinematicDataUnitInterface* ku, const std::vector<std::string>& v)
        {
            return ku->getJointAngles(v);
        }
        static std::string getControlMode()
        {
            return ControlModes::PositionMode;
        }
    };

    template <>
    struct PassThroughControllerTargetTypeTraits<PlatformWheelVelocityTarget>
    {
        static float* getTargetDatamember(PlatformWheelVelocityTarget& t)
        {
            return  &t.velocity;
        }
        static const float* getFrontRightWheelVelocity(const PlatformDataUnitInterface* pu)
        {
            return pu->getFrontRightWheelVelocity();
        }
        static const float* getFrontLeftWheelVelocity(const PlatformDataUnitInterface* pu)
        {
            return pu->getFrontLeftWheelVelocity();
        }
        static const float* getRearRightWheelVelocity(const PlatformDataUnitInterface* pu)
        {
            return pu->getRearRightWheelVelocity();
        }
        static const float* getRearLeftWheelVelocity(const PlatformDataUnitInterface* pu)
        {
            return pu->getRearLeftWheelVelocity();
        }
        static  std::string getControlMode()
        {
            return ControlModes::VelocityMode;
        }
    };

    LVL1ControllerRegistration<PassThroughController<JointPositionTarget>> registrationSomeControllerPositionPassThroughController("PositionPassThroughController");
    LVL1ControllerRegistration<PassThroughController<JointVelocityTarget>> registrationSomeControllerVelocityPassThroughController("VelocityPassThroughController");
    LVL1ControllerRegistration<PassThroughController<JointTorqueTarget  >> registrationSomeControllerJointPassThroughController("TorquePassThroughController");

    template <typename TargetType>
    std::string PassThroughController<TargetType>::getClassName(const Ice::Current&) const
    {
        return "PassThroughController_" + Traits::getControlMode();
    }

    template <typename TargetType>
    PassThroughController<TargetType>::PassThroughController(LVL1ControllerDataProviderInterfacePtr prov, LVL1ControllerConfigPtr config)
    {
        PassThroughControllerConfigPtr cfg = PassThroughControllerConfigPtr::dynamicCast(config);
        ARMARX_CHECK_EXPRESSION_W_HINT(cfg,
                                       "The provided config has the wrong type! The type is " << config->ice_id()
                                       << " instead of " << PassThroughControllerConfig::ice_staticId());

        //make sure the used units are supported
        auto kinunit = prov->getRTKinematicDataUnit();
        ARMARX_CHECK_EXPRESSION_W_HINT(kinunit, "The DataProvider " << prov->getName() << " has no kinematic data unit");
        //get pointers to sensor values from units
        ARMARX_INFO << "PassThroughController for " << cfg->jointNames;
        jointStates = Traits::getSensorValues(kinunit, cfg->jointNames) ;

        //not initialized, this is done in rtPreActivateController
        iceTargets.resize(cfg->jointNames.size(), 0);

        //get pointers for the results of this controller
        lvl0Targets.reserve(cfg->jointNames.size());
        for (const auto & j : cfg->jointNames)
        {
            auto target = dynamic_cast<TargetType*>(prov->getJointTarget(j, Traits::getControlMode()));
            ARMARX_CHECK_EXPRESSION_W_HINT(target, "The joint " << j << " has no controll mode " << Traits::getControlMode());
            lvl0Targets.emplace_back(Traits::getTargetDatamember(*target));
        }
        indices = toIndexMap(cfg->jointNames);
    }

    template <typename TargetType>
    void PassThroughController<TargetType>::rtRun(const IceUtil::Time&, const IceUtil::Time&)
    {
        for (std::size_t i = 0; i < iceTargets.size(); ++i)
        {
            *lvl0Targets.at(i) = iceTargets.at(i).val.load();
        }
    }

    template <typename TargetType>
    void PassThroughController<TargetType>::rtSwapBufferAndRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
    {
        rtRun(sensorValuesTimestamp, timeSinceLastIteration);
    }

    template <typename TargetType>
    void PassThroughController<TargetType>::rtPreActivateController()
    {
        for (std::size_t i = 0; i < jointStates.size(); ++i)
        {
            iceTargets.at(i).val.store(*jointStates.at(i));
            //set it to zero
            iceTargets.at(i).val.store(0.0);
        }
    }

    template <typename TargetType>
    void PassThroughController<TargetType>::setJoint(const std::string& name, Ice::Float value, const Ice::Current&)
    {
        auto targetIt = indices.find(name);
        if (targetIt == indices.end())
        {
            throw InvalidArgumentException {"The joint " + name + " is not controlled by this (" + getName() + ") controller"};
        }
        iceTargets.at(targetIt->second).val.store(value);
    }

    template <typename TargetType>
    void PassThroughController<TargetType>::setJoints(const StringFloatDictionary& values, const Ice::Current&)
    {
        for (const auto & value : values)
        {
            setJoint(value.first, value.second);
        }
    }

}
#endif
