/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::KinematicDataUnit
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_LIB_RobotAPI_KinematicDataUnit_H
#define _ARMARX_LIB_RobotAPI_KinematicDataUnit_H

#include <vector>
#include <string>
#include <unordered_map>
#include <algorithm>

#include <ArmarXCore/core/util/algorithm.h>

namespace armarx
{
    class KinematicDataUnitInterface
    {
    public:
        virtual const std::vector<std::string>& getJointNames() const = 0;

        virtual std::vector<const float*> getJointAngles(const std::vector<std::string>& joints) const = 0;

        virtual std::vector<const float*> getJointVelocities(const std::vector<std::string>& joints) const = 0;

        virtual std::vector<const float*> getJointTorques(const std::vector<std::string>& joints) const = 0;

        virtual std::vector<const float*> getJointCurrents(const std::vector<std::string>& joints) const = 0;

        virtual std::vector<const float*> getJointMotorTemperatures(const std::vector<std::string>& joints) const = 0;

        virtual std::vector<const float*> getJointGearTemperatures(const std::vector<std::string>& joints) const = 0;

        virtual std::vector<const int32_t*>
        getJointRawPositionValues(const std::vector<std::string>& joints) const = 0;

        virtual std::vector<const int32_t*>
        getJointRawVelocityValues(const std::vector<std::string>& joints) const = 0;

        virtual std::vector<const int16_t*> getJointRawCurrentValues(const std::vector<std::string>& joints) const = 0;

        virtual std::vector<const int32_t*> getJointRawTorqueValues(const std::vector<std::string>& joints) const = 0;

        virtual std::vector<const int32_t*>
        getJointRawGearTemperatures(const std::vector<std::string>& joints) const = 0;

        virtual std::vector<const int32_t*>
        getJointRawAbsPositionValues(const std::vector<std::string>& joints) const = 0;

        virtual std::vector<const int16_t*>
        getJointRawMotorTemperatureValues(const std::vector<std::string>& joints) const = 0;
    };

    class KinematicDataUnitPtrProvider : virtual public KinematicDataUnitInterface
    {
    public:
        KinematicDataUnitPtrProvider(const std::vector<std::string>& jointNames) :
            jointAngles(jointNames.size(), nullptr),
            jointVelocites(jointNames.size(), nullptr),
            jointTorques(jointNames.size(), nullptr),
            jointCurrents(jointNames.size(), nullptr),
            jointMotorTemperatures(jointNames.size(), nullptr),
            jointGearTemperatures(jointNames.size(), nullptr),
            jointAbsPositions(jointNames.size(), nullptr),
            jointRawPositionValues(jointNames.size(), nullptr),
            jointRawVelocityValues(jointNames.size(), nullptr),
            jointRawCurrentValues(jointNames.size(), nullptr),
            jointRawTorqueValues(jointNames.size(), nullptr),
            jointRawGearTemperatures(jointNames.size(), nullptr),
            jointRawAbsPositionValues(jointNames.size(), nullptr),
            jointRawMotorTemperatureValues(jointNames.size(), nullptr),
            jointNames {jointNames},
        jointIndices {toIndexMap(jointNames)} {}

        KinematicDataUnitPtrProvider() = default;

        KinematicDataUnitPtrProvider(const KinematicDataUnitPtrProvider&) = default;

        void setJointNames(std::vector<std::string>& names)
        {
            jointNames = names;
            jointIndices = toIndexMap(names);

            jointAngles = std::vector<float*>(names.size(), nullptr);
            jointVelocites = std::vector<float*>(jointNames.size(), nullptr);
            jointTorques = std::vector<float*>(jointNames.size(), nullptr);
            jointCurrents = std::vector<float*>(jointNames.size(), nullptr);
            jointMotorTemperatures = std::vector<float*>(jointNames.size(), nullptr);
            jointGearTemperatures = std::vector<float*>(jointNames.size(), nullptr);
            jointAbsPositions = std::vector<float*>(jointNames.size(), nullptr);
        }

        const std::vector<std::string>& getJointNames() const
        {
            return jointNames;
        }

        std::size_t getJointIndex(const std::string& joint) const
        {
            return jointIndices.at(joint);
        }

        virtual std::vector<const float*> getJointAngles(const std::vector<std::string>& joints) const
        {
            return getPointers(joints, jointAngles);
        }

        virtual std::vector<const float*> getJointVelocities(const std::vector<std::string>& joints) const
        {
            return getPointers(joints, jointVelocites);
        }

        virtual std::vector<const float*> getJointTorques(const std::vector<std::string>& joints) const
        {
            return getPointers(joints, jointTorques);
        }

        virtual std::vector<const float*> getJointCurrents(const std::vector<std::string>& joints) const
        {
            return getPointers(joints, jointCurrents);
        }

        virtual std::vector<const float*> getJointMotorTemperatures(const std::vector<std::string>& joints) const
        {
            return getPointers(joints, jointMotorTemperatures);
        }

        virtual std::vector<const float*> getJointGearTemperatures(const std::vector<std::string>& joints) const
        {
            return getPointers(joints, jointGearTemperatures);
        }

        virtual std::vector<const float*> getJointAbsPositions(const std::vector<std::string>& joints) const
        {
            return getPointers(joints, jointAbsPositions);
        }

        virtual std::vector<const int32_t*> getJointRawPositionValues(const std::vector<std::string>& joints) const
        {
            return getPointers(joints, jointRawPositionValues);
        }

        virtual std::vector<const int32_t*> getJointRawVelocityValues(const std::vector<std::string>& joints) const
        {
            return getPointers(joints, jointRawVelocityValues);
        }

        virtual std::vector<const int16_t*> getJointRawCurrentValues(const std::vector<std::string>& joints) const
        {
            return getPointers(joints, jointRawCurrentValues);
        }

        virtual std::vector<const int32_t*> getJointRawTorqueValues(const std::vector<std::string>& joints) const
        {
            return getPointers(joints, jointRawTorqueValues);
        }

        virtual std::vector<const int32_t*> getJointRawGearTemperatures(const std::vector<std::string>& joints) const
        {
            return getPointers(joints, jointRawGearTemperatures);
        }

        virtual std::vector<const int32_t*>
        getJointRawAbsPositionValues(const std::vector<std::string>& joints) const
        {
            return getPointers(joints, jointRawAbsPositionValues);
        }

        virtual std::vector<const int16_t*>
        getJointRawMotorTemperatureValues(const std::vector<std::string>& joints) const
        {
            return getPointers(joints, jointRawMotorTemperatureValues);
        }


    protected:
        std::vector<float*> jointAngles;
        std::vector<float*> jointVelocites;
        std::vector<float*> jointTorques;
        std::vector<float*> jointCurrents;
        std::vector<float*> jointMotorTemperatures;
        std::vector<float*> jointGearTemperatures;
        std::vector<float*> jointAbsPositions;

        std::vector<int32_t*> jointRawPositionValues;
        std::vector<int32_t*> jointRawVelocityValues;
        std::vector<int16_t*> jointRawCurrentValues;
        std::vector<int32_t*> jointRawTorqueValues;
        std::vector<int32_t*> jointRawGearTemperatures;
        std::vector<int32_t*> jointRawAbsPositionValues;
        std::vector<int16_t*> jointRawMotorTemperatureValues;
    private:
        template<class T>
        std::vector<const T*>
        getPointers(const std::vector<std::string>& joints, const std::vector<T*>& targets) const
        {
            std::vector<const T*> result;
            result.reserve(joints.size());
            std::transform(
                joints.begin(), joints.end(), std::back_inserter(result),
                [targets, this](const std::string & joint)
            {
                return targets.at(getJointIndex(joint));
            }
            );
            return result;
        }

        std::vector<std::string> jointNames;
        std::unordered_map<std::string, std::size_t> jointIndices;
    };
}

#endif
